const logger = require('../../common.js').createLogger();
const utils = require('../../utils');

const NeutrinoExternalInterface = function(neutrino, key, gamestate) {
  this.getId = function() {
    return key;
  };
  this.awardScore = function(playerId, score, cb) {
    logger.trace('award score called');
    neutrino.awardScore(key, playerId, score);
    if (cb) cb();
  };
  this.endGame = function(endGameReason, cb) {
    logger.info('External plugin called for ending the game %s', key);
    neutrino.endGame(key, endGameReason, cb);
  };
  this.gameObjectByIds = function(...ids) {
    let objs = {};
    for (let id of ids) {
      objs[id] = utils.gameObjectToExtGameObject(gamestate.gameObjectById(id));
    }
    return objs;
  };
  this.changeOwner = function(playerId, ...ids) {
    for (let id of ids) {
      let obj = gamestate.gameObjectById(id);
      obj.owner = playerId;
    }
  };

  this.getNextTurnPlayerId = function() {
    let playerInfo = gamestate.getNextPlayer();
    if (playerInfo) return playerInfo.playerId;
  };

  this.getCurrentTurnPlayerId = function() {
    let playerInfo = gamestate.currentTurn();
    if (playerInfo) return playerInfo.playerId;
  };

  this.getCurrentTurnIndex = function() {
    return gamestate.getCurrentTurnIndex();
  };

  this.getAllObjIds = function() {
    return Object.keys(gamestate.gameObjects);
  };

  this.setGameProp = function(key, value) {
    gamestate.setProp(key, value);
  };

  this.getGameProp = function(key) {
    return gamestate.getProp(key);
  };

  this.setPlayerProp = function(playerId, key, value) {
    let player = gamestate.getPlayerByPlayerId(playerId);
    if (!player) return;
    return player.setProp(key, value);
  };

  this.getPlayerProp = function(playerId, key) {
    let player = gamestate.getPlayerByPlayerId(playerId);
    if (!player) return;
    return player.getProp(key);
  };

  this.getPlayerScore = function(playerId) {
    let player = gamestate.getPlayerByPlayerId(playerId);
    if (!player) return;
    return player.score;
  };

  this.getAllPlayerIds = function() {
    let playerIds = [];
    for (let player of gamestate.players) playerIds.push(player.playerId);
    return playerIds;
  };

  /**
   * Initializes collision map from redis for collision detection
   * @param cb of form cb(err, result)
   */
  this.initCollisionMap = function(cb) {
    neutrino.initCollision(gamestate, cb);
  };

  /**
   * First persists all game objects and then
   * updates all game objects, that have position changed, position into
   * the collision map
   * @param cb
   */
  this.updateCollisionMap = function(cb) {
    neutrino.saveGameState(gamestate, (err, res) => {
      if (err && typeof cb === 'function') return cb(err, null);
      neutrino.updateCollisionObjs(gamestate, cb);
    });
  };

  /**
   * returns booleans if objIdB is supplied, or else return an array of objects
   * collided with objIdA
   */
  this.collidedObjsWithObject = function(objIdA, objIdB, cb) {
    neutrino.collidedObjsWithObject(
        gamestate.getRedisKey(),
        objIdA,
        objIdB,
        cb,
    );
  };

  /**
   * returns list of objs colliding with the boundary
   * @param boundary should be of format {minX: 0, minY: 1, maxX: 2, maxY:3}
   * @param cb callback to be called after the result be found, of format cb(err, result)
   */
  this.collidedObjsWithBoundary = function(boundary, cb) {
    neutrino.collidedObjsWithBoundary(gamestate.getRedisKey(), boundary, cb);
  };

  this.syncGameStateServVars = function() {
    logger.trace('Calling external index to send server vars');
    neutrino.sendGameStateServVars(key);
  };

  this.syncPlayerServVars = function(playerId) {
    neutrino.sendPlayerServVars(key, playerId);
  };

  this.syncGameObjServVars = function(...objIds) {
    neutrino.sendGameObjServVars(key, ...objIds);
  };

  this.forceUpdateGameObjects = function(...objTransList) {
    logger.trace('Sending force update for game objects', objTransList);
    neutrino.forceUpdateGameObjects(key, ...objTransList);
  };
  this.addTransient = function(key, val) {
    gamestate.addTransient(key, val);
  };
  this.removeTransient = function(key) {
    gamestate.removeTransient(key);
  };
  this.getTransient = function(key) {
    return gamestate.getTransient(key);
  };
  this.nextTurn = function(turnData) {
    neutrino.nextTurn(key, turnData);
  };
  this.continueTurn = function(turnData) {
    neutrino.continueTurn(key, turnData);
  };
  this.save = function(cb) {
    neutrino.saveGameState(gamestate, (err, res) => {
      if (err) return cb(err);
      cb(null);
    });
  };
  this.getCurrentTurnExpiry = function() {
    let durationProp = gamestate.getProp('turnExpiry');
    if (!durationProp) return -1;
    return parseInt(durationProp);
  };
  this.issueSavepoint = function(savepointName, cb) {
    neutrino.issueSavePoint(key, savepointName, cb);
  };
  this.restoreFromSavepoint = function(savepointname, cb) {
    neutrino.restoreFromSavePoint(key, savepointname, cb);
  };
  this.getState = function() {
    return gamestate.state;
  };
  this.spawnObjects = function(objectList, isImmediate, targetPlayerId) {
    neutrino.spawnObjects(key, objectList, isImmediate, targetPlayerId);
  };
  this.getConnectedPlayers = function() {
    return neutrino.getConnectedPlayers(key);
  };
  this.getDuration = function() {
    return gamestate.duration || -1;
  };
  this.getScheduledStartTime = function() {
    return gamestate.scheduledStartTime;
  };

  this.gameObjectPositionByIds = function(...ids) {
    let objs = {};
    for (let id of ids) {
      let obj = gamestate.gameObjectById(id);
      objs[id] = {x: obj.position.x, y: obj.position.y, z: obj.position.z};
    }
    return objs;
  };

  this.gameObjectForceByIds = function(...ids) {
    let objs = {};
    for (let id of ids) {
      let obj = gamestate.gameObjectById(id);
      objs[id] = {x: obj.force.x, y: obj.force.y, z: obj.force.z};
    }
    return objs;
  };

  this.getGameMeta = function() {
    let gameMeta = neutrino.registeredGames[gamestate.gameType].meta;
    return utils.gameMetaToExtGameMeta(gameMeta);
  };

  this.forfeitGameForPlayer = function(playerId, reason) {
    neutrino.forfeitGame(key, playerId, reason);
  };

  this.abandonGame = function(reason) {
    neutrino.abandonGame(key, reason);
  };
  this.getGameStateTimeElapsed = function() {
    return gamestate.elapsedTimeMillis;
  };
  this.forceUpdateForceGameObjects = function(toSend, ...objForceList) {
    logger.trace({key: key}, 'Force updating force for gameObjects',
        stringify(objForceList));
    if (toSend)
      return logger.error({key: key},
          'Force update of force to clients is not implemented, dropping it');
    for (let obj of objForceList) {
      let gameObj = gamestate.gameObjects[obj.id];
      if (obj.force.x != null) gameObj.force.x = obj.force.x;
      if (obj.force.y != null) gameObj.force.y = obj.force.y;
      if (obj.force.z != null) gameObj.force.z = obj.force.z;
    }
  };

  this.forceUpdateSyncVars = function(toSend, ...objSyncVarsList) {
    if (logger.trace())
      logger.trace({key: key}, 'Force updating sync vars for gameObjects',
          stringify(objSyncVarsList));
    if (toSend)
      return logger.error({key: key},
          'Force update of sync vars to clients is not implemented, dropping it');
    for (let obj of objSyncVarsList) {
      let gameObj = gamestate.gameObjects[obj.id];
      for (let [key, value] of Object.entries(obj.syncVars))
        gameObj.setSyncVar(key, value);
    }
  };
  this.setProp = function(key, value) {

  };
  this.getProp = function(key) {

  };
  this.turnDuration = 30000;
  this.allowedIdleTurnsCount = 3;
  this.timeElapsed = 5;
};

module.exports = NeutrinoExternalInterface;
